unit uKC;

interface

uses SysUtils, Windows;

const
   PwdMask: Array[0..7] of Byte = ($01, $16, $13, $05, $10, $02, $17, $04);

   procedure UnMaskPassword(var Pass: Array of Byte);
   procedure MaskPassword(var Pass: Array of Byte);
   function IsValid(Pass: Array of Byte): Boolean;
   procedure KCtoASCII(var Pass: Array of Byte);
   function GeneratePassword(Unused: Byte; Zone: Byte; Level: Byte): String;
   procedure Encrypt(Value:Byte; m16: Byte; XR: Byte; YR: Byte; var Pass: Array of Byte);
implementation

procedure Encrypt(Value: Byte; m16: Byte; XR: Byte; YR: Byte; var Pass: Array of Byte);
var
   CR: Byte;
begin
     while (YR > 0) do begin
           CR := (Value and $80);
           if CR = $80 then CR := 1;
           Value := Value shl 1;
           Pass[XR] := ((Pass[XR] shl 1) or CR);
           Dec(m16);

           if m16 = 0 then begin
              Inc(XR);
              m16 := 5;
           end;

           Dec(YR);
     end;
end;

procedure UnMaskPassword(var Pass: Array of Byte);
var
   i, PWDIndex: Integer;
begin
    PWDIndex := (Pass[5] and 1) * 4;
    for i := 3 downto 0 do begin
        Pass[i] := ((Pass[i] - PwdMask[PWDIndex]) and $1F);
        Inc(PWDIndex);
    end;
end;

procedure MaskPassword(var Pass: Array of Byte);
var
   i, PWDIndex: Integer;
begin
    PWDIndex := (Pass[5] and 1) * 4;
    for i := 3 downto 0 do begin
        Pass[i] := ((Pass[i] + PwdMask[PWDIndex]) and $1F);
        Inc(PWDIndex);
    end;
end;

function IsValid(Pass: Array of Byte): Boolean;
var
   x, y: Integer;
begin

    UnMaskPassword(Pass);

    x := (Pass[0] xor Pass[1] xor Pass[2] xor Pass[3]);
    if (x <> Pass[4]) then begin
       Result := False;
       Exit;
    end;

    x := (Pass[0] + Pass[1] + Pass[2] + Pass[3]);
    y := x shr 5;
    x := x and $1F;
    if ((x <> Pass[5]) or (y <> Pass[6])) then begin
       Result := False;
       Exit;
    end;

    y := Pass[5];
    while(y >= 0) do Dec(y, 7);
    Inc(y, 7);
    x := (Pass[y] xor $1F);
    if (x <> Pass[7]) then begin
       Result := False;
       Exit;
    end;

    Result := True;
end;

procedure KCtoASCII(var Pass: Array of Byte);
var
   i: Integer;
begin

  for i := 0 to 7 do begin
    if ((Pass[i] >= 0) and (Pass[i] <= $19)) then begin
       Inc(Pass[i], $41);
    end;
    if (Pass[i] = $1A) then Pass[i] := Ord('a');
    if (Pass[i] = $1B) then Pass[i] := Ord('b');
    if (Pass[i] = $1C) then Pass[i] := Ord('d');
    if (Pass[i] = $1D) then Pass[i] := Ord('e');
    if (Pass[i] = $1E) then Pass[i] := Ord('g');
    if (Pass[i] = $1F) then Pass[i] := Ord('h');
  end;
end;

function GeneratePassword(Unused: Byte; Zone: Byte; Level: Byte): String;
var
   Pass: Array[0..7] of Byte;
   i: Integer;
begin
     // This must be the most awful way to clear an array.
     Pass[0] := 0;
     Pass[1] := 0;
     Pass[2] := 0;
     Pass[3] := 0;
     Pass[4] := 0;
     Pass[5] := 0;
     Pass[6] := 0;
     Pass[7] := 0;

     // Encrypt the unused byte (used in the japanese version though)
     Unused := Unused shl 2;
     Encrypt(Unused, 5, 0, 6, Pass);

     // Encrypt the Zone
     Zone := Zone shl 5;
     Encrypt(Zone, 4, 1, 3, Pass);

     // Encrypt the Level
     Level := Level shl 2;
     Encrypt(Level, 1, 1, 6, Pass);

     // Apply a XOR mask
     for i := 0 to 3 do begin
         Pass[i + 1] := Pass[i] xor Pass[i + 1];
     end;

     // Control Code #1
     Pass[4] := Pass[0] xor Pass[1] xor Pass[2] xor Pass[3];

     // Control Code #2 & #3
     Pass[6] := ((Pass[0] + Pass[1] + Pass[2] + Pass[3]) shr 5);
     Pass[5] := ((Pass[0] + Pass[1] + Pass[2] + Pass[3]) and $1F);

     // Control Code #4
     i := Pass[5];
     while(i >= 0) do Dec(i, 7);
     Inc(i, 7);
     Pass[7] := (Pass[i] xor $1F);

     // Mask the password
     MaskPassword(Pass);

     // Convert to ASCII
     KCtoASCII(Pass);

     Result := Chr(Pass[0]) + Chr(Pass[1]) +
               Chr(Pass[2]) + Chr(Pass[3]) +
               Chr(Pass[4]) + Chr(Pass[5]) +
               Chr(Pass[6]) + Chr(Pass[7]);
end;

end.
